[INHERIT('sys$share:starlet'
	,'sys$share:pascal$lib_routines'
	,'sys$share:pascal$str_routines'
	)]
PROGRAM HTML_BOOK(input,output);

(******************************************************************************
PROGRAM DESCRIPTION:

 Bookreader-to-HTML gateway

AUTHORS:

 Stellan Lagerstrm		<stellanl@e.kth.se>

 Parts of the code come originally from the DECUS VTBOOK project,
 and thus have been written by one or more of
 Cris Sharman, Darrell Burkhead, Hunter Goatley, Curt Snyder, and Bret D Wortman.

 The bugs are probably all mine, though...
 
MODIFICATION HISTORY:

Ver  Date	Name    	Description 
0.1  1995-05-12	STELLANL	First working version, no figures
0.2  1995-05-23	STELLANL	Figure bitmaps output as XBM files
0.3  1995-06-02	STELLANL	Figure bitmaps as PBM files, piped to GIF-converter
0.4  1995-06-12	STELLANL	Figure text (decwrite books) support added
0.5  1995-06-19	STELLANL	Fix comment syntax, start improving <PRE>
0.6  1995-08-31	STELLANL	Fix bugs in previous version. <PRE> still not done.
0.6A 1995-08-31 STELLANL	Use BOOK_FIGURE logical more. Longer output buffer.
0.7  1995-09-10 STELLANL	Escape <&> and kill hyphens.
0.8  1995-09-11 STELLANL	Use BOOK_FIGURE in URLs too; do more with <PRE>.

STILL TO BE DONE:
 *  Detecting and generating <UL> lists.
 *  Better Symbol and dectech font character handling
 *  Batch-generating a whole book.
 *  Eliminating the .COM wrapper script.
 *  Making the figure converter subprocess time out if hung.
 *  Detecting tables for <PRE>, or maybe
 *  giving the option of clicking a <PRE>link to the same page: "?PRE"
 ******************************************************************************)

CONST
  htbook_version =  'HTBOOK 0.8';

(*  html		    = 1;(* approx. (mosaic 2.4) *)
  html		    = 2;	(* approx. (mosaic 2.4) *)
(*  html		    = 3;	(* HTML v3 *)
  max_index	    = 10;
  nbsp		    = CHR(160);
  average_charwidth = 0.25; (*reasonable guess*)
  min_scale	    = 10.0;
  letters	    = ['A'..'Z','a'..'z',
		       CHR(192)..CHR(214),CHR(216)..CHR(246),CHR(248)..CHR(255)
		      ];
  crlf=CHR(13)+CHR(10);
  
TYPE	 
  $ubyte = [byte] 0..255;
  $uword = [word] 0..65535;
  $uquad = [quad,unsafe] RECORD L0,L1:UNSIGNED; END;
  $bool	 = [bit(1),unsafe] BOOLEAN;
  bit8	 = [byte] PACKED ARRAY[0..7] OF $bool;

  recbuff   = PACKED ARRAY [1..32768] OF CHAR;
  record_p  = ^recbuff;
  pun	    = [unsafe]^UNSIGNED;
  puw	    = [unsafe]^$uword;
  pub	    = [unsafe]^$ubyte;
  rfa_array = [word(3)]ARRAY [1..3] of $UWORD;
  
  br_sectptr = RECORD
    rfa	: rfa_array;
    offset, vpos, partno : INTEGER;
  END;				      

  encodings =(latin1,symbol,mathx,maths,mathi,other);
  
  rectyp = (type_unknown, part_first, unknown2, part_body, part_index, last, part_secmap,
	    type_f_desc, unknown8, part_font, continue_mid, continue_end,
	    sb_tblhdr, asciz_32, sb_font, sb_fdesc, sb_ixtxt, unknown11,
	    sb_bodytext, sb_figure, sb_hotspot, sb_unknown15, sb_extension,
	    unknown17, sb_license);

  rectyp_w = [WORD] rectyp;
  
  br_record = PACKED RECORD
    rtype: rectyp_w;
    len: INTEGER;
  END;
  br_record_p = ^br_record;
  
  ascic = PACKED RECORD
    len: $ubyte;
    txt: PACKED ARRAY [1..255] OF CHAR;
  END;
  
  br_txt_rec = PACKED RECORD
    subsectyp, len: $ubyte;
    hor, ver: $uword;
    fontno, x, y: $ubyte;
    txt: ascic;
  END;
  br_txt_rec_p = ^br_txt_rec;

  br_key_head =	PACKED RECORD
    subtype : [pos(0)]rectyp_w;
    sublen  : [pos(16)]INTEGER;
    CASE (*subtype*) rectyp OF
      sb_tblhdr:( (* on first page; table header*)
		 unknown0: INTEGER; (* total record size *)
		 keyid: $ubyte; (* 5=contents, x0a=index, 0=other *)
		 count: INTEGER; (*count of entries*)
		 part: INTEGER; (*partnumber*)
		 title_b_len: $ubyte;
		 title_asciz: PACKED ARRAY [1..255] OF CHAR);
      sb_ixtxt:(ixhdr: ARRAY [1..7] OF $uword;
	        ixtxt: br_txt_rec);
      sb_bodytext:(bodhdr: ARRAY [2..10] OF INTEGER;
	           bodtxt: br_txt_rec);
      sb_hotspot:(hot_unk1, hot_unk2: INTEGER;
		  hot_sect, horpos, verpos, len, wid, hot_target: INTEGER;
		  hot_unk3: INTEGER);
      sb_extension: (ext_unk1,ext_unk2,ext_sect, (* extension "grayspot" on body pages*)
		     ext_x,ext_y,ext_w,ext_h,
		     ext_unk3,ext_unk4:INTEGER);
      sb_license:(alt_license: PACKED ARRAY [1..255] OF CHAR);	(* on first page*)
      sb_figure:(fig_unk1, fig_unk2:	 INTEGER;
		 fig_sect,
		 fig_x,fig_y,fig_w,fig_h:INTEGER;
		 fig_unk3, fig_picbytes: INTEGER;
		 fig_primitive,fig_unk5, fig_unk6, fig_unk7:	 $ubyte;
		 CASE BOOLEAN OF
		   TRUE:(fig_columns, fig_rows:	 $uword;
		         fig_pixels:		 PACKED ARRAY [0..65535] OF bit8);
                   FALSE:(figtxt_font:$uword;
			  figtxt_unk8:$uword;
			  figtxt_data:VARYING[1024] OF CHAR));
    END;
  pbr_key_head = ^br_key_head;
  
  br_first_record = PACKED RECORD
    rtype	  : rectyp_w;
    len		  : INTEGER;
    major_version : $uword;	(* ODS major version number *)
    minor_version : $uword;	(* ODS eco level *)
    author_asciz  : PACKED ARRAY [1..24] OF CHAR;
    product_asciz : PACKED ARRAY [1..36] OF CHAR;
    partcount	  : INTEGER;		(* Number of parts *)
    sectioncount  : INTEGER;		(* Number of subparts *)
    numdirs,				(* Number of directories *)
    numfonts,				(* Number of fonts *)
    maxfontindex  : INTEGER;		(* Highest font id *)
    lastptr	  : rfa_array;		(* RFA of part index record (last) *)
    lastsiz	  : $uword;		(* Size of part index record(s) *) 
    docverspage,			(* page id of the book header ext. *)
    index_page,				(* page id of chunk index *)
    title_page,				(* page id of chunk titles *)
    sym_table,				(* page id of symbol table *)
    font_page,				(* page id of font definitions *)
    start_here,				(* id of page to open with book *)
    copyrght	  : INTEGER;		(* id of copyright chunk *)
    title_b_len	  : $ubyte;
    title_asciz	  : PACKED ARRAY [1..128] OF CHAR;
    credate	  : $uquad;
    id_asciz	  : PACKED ARRAY [1..33] OF CHAR;
    (*Key headers follow here*)
  END;

  br_a32 = PACKED ARRAY [1..32] OF CHAR;
  pbr_a32 = ^br_a32;
  
  br_part_ptr =	RECORD
    subtype: rectyp_w;
    len: INTEGER;
    partno: ARRAY [0..255] OF UNSIGNED;
  END;
  
  bodypart_hdr = PACKED RECORD
    partno      : INTEGER;
    unknown_one : INTEGER;
    sectcnt     : INTEGER;
    prevpart    : INTEGER;
    nextpart    : INTEGER;
  END;
  
  br_fdesc = PACKED RECORD
    subtype: rectyp_w;
    len: INTEGER;
    txt: PACKED ARRAY [1..255] OF CHAR;
  END;
  pbr_fdesc = ^br_fdesc;
    
  br_ptr = PACKED RECORD
    rfa: rfa_array;
    siz: INTEGER;
  END;
  
  br_last_record = PACKED RECORD
    rtype : rectyp_w;
    len   : INTEGER; (*x1fa - no idea why*)
    part  : ARRAY [0..10000] OF br_ptr;
  END;    

  part(n: INTEGER) = ARRAY [0..n - 2] OF br_ptr;
  sect(n: INTEGER) = ARRAY [0..n] OF br_sectptr;

  br_font = PACKED RECORD	(* font data in book file *)
    subtype : rectyp_w;
    len	    : INTEGER;
    fontno  : $uword;
    txt	    : PACKED ARRAY [1..255] OF CHAR;
  END;	    
  pbr_font = ^br_font;

  fontdata = RECORD		(* remembered font data *)
    charsize: INTEGER;
    in_bold:BOOLEAN;
    in_italics:BOOLEAN;
    underline:BOOLEAN;
    encoding:encodings;
    render: UNSIGNED;	
    name:STRING(80);
  END;

  hotspotp = ^hotspot;
  hotspot = RECORD		(* remembered hotspot *)
    pnxt, pprv: hotspotp;
    s,x,y,w,h,target: INTEGER;
    gray:BOOLEAN;
  END;

  keymenu(n: INTEGER) = RECORD
    sect: ARRAY [1..n] OF INTEGER;
    name: ARRAY [1..n] OF VARYING [132] OF CHAR;
  END;

  itmlst = RECORD
    buflen : $UWORD;
    itmcod : $UWORD;
    bufadr : UNSIGNED;
    retlen : UNSIGNED;
    null   : UNSIGNED;
  END;
	
CONST
  subpart_minlen = BYTE_OFFSET(br_key_head, sublen) + 4;
  
VAR
  debug	       : INTEGER VALUE 0;
  debug_font   : BOOLEAN VALUE FALSE;(* include font info comments *)
  debug_secmap : BOOLEAN VALUE FALSE;(* include section map comments *)
  debug_spot   : BOOLEAN VALUE FALSE;(* include hotspot/greyspot comments *)
  book_url     : STRING(254) VALUE 'URL';
  book_id      : STRING(255);

  parts: ^part;
  sects: ^sect;
  fonts: ARRAY [0..300] OF fontdata; (* should be dynamic... *)
  
  cury,curx, maxfont: INTEGER;
  curstr:STRING(2046) VALUE '';
  hyphen:BOOLEAN VALUE FALSE;
  hotspots, curhotspot: hotspotp;
  
  cursect    : INTEGER;
  curfont    : INTEGER;
  scale	     : REAL;
  post	     : STRING(40) VALUE '';
  
  nkeys: INTEGER VALUE 0;
  keypart, keykind, keycount: ARRAY [1..max_index] OF INTEGER;
  keyname: ARRAY [1..max_index] OF VARYING [30] OF CHAR;
  
  broken_subpart: ^STRING;
  
  net_block_count: UNSIGNED;
  xab_itmlst: itmlst VALUE [buflen: 4; itmcod: xab$_net_block_count;
  OTHERWISE ZERO];
  
  rab: rab$type VALUE [rab$b_bid: rab$c_bid; rab$b_bln: rab$c_bln;
		       rab$b_mbc: 127; rab$v_rah: TRUE; rab$v_loc: TRUE;
		       OTHERWISE ZERO];
  fab: fab$type VALUE [fab$b_bid: fab$c_bid; fab$b_bln: fab$c_bln;
		       fab$b_rtv: - 1; OTHERWISE ZERO];
  xab: xab$type VALUE [xab$b_cod: xab$c_fhc; xab$b_bln: xab$c_fhclen;
		       OTHERWISE ZERO];
  xabitm: xab$type VALUE [xab$b_cod: xab$c_itm; xab$b_bln: xab$c_itmlen;
                          xab$b_mode: xab$k_setmode; OTHERWISE ZERO];


PROCEDURE ASSERT(s:INTEGER);
BEGIN
  IF NOT ODD(s) THEN LIB$STOP(s);
END;


PROCEDURE INIT_RMS_BLK(VAR fab: fab$type;
		       VAR rab: rab$type);
BEGIN
  xab_itmlst.bufadr := IADDRESS(net_block_count);
  xabitm.xab$l_itmlst := IADDRESS(xab_itmlst);
  xab.xab$l_nxt := IADDRESS(xabitm);
  fab.fab$l_xab := IADDRESS(xab);
  rab.rab$l_fab := IADDRESS(fab);
END;



FUNCTION CHECK_LICENSE(author: PACKED ARRAY [l1..u1: INTEGER] OF CHAR;
		       product: PACKED ARRAY [l2..u2: INTEGER] OF CHAR;
		       maxlen: [TRUNCATE] INTEGER): BOOLEAN;
VAR
  i, j : INTEGER;
  s    : VARYING [80] OF CHAR;
  chk  : BOOLEAN;
       
BEGIN
  j := INDEX(product, CHR(0)) - 1;

  IF (j < 0) THEN j := LENGTH(product);

  IF PRESENT(maxlen) AND_THEN (j > maxlen) THEN j := maxlen;

  IF (j = 0) THEN BEGIN

    IF (debug>0) THEN WRITELN('<!-- No license required -->'+crlf);

    check_license := TRUE;
  END
  ELSE BEGIN
    i := INDEX(author, CHR(0)) - 1;

    IF (i < 0) THEN i := LENGTH(author);

    s := 'LMF$' + SUBSTR(author, 1, i) + '_' + SUBSTR(product, 1, j);
    chk := ODD($TRNLNM(tabnam := 'LMF$LICENSE_TABLE',
	       attr := lnm$m_case_blind,
	       lognam := s));

    IF (debug>0) THEN
      IF chk
	THEN WRITELN('<!-- License ', s, ' present. -->'+crlf)
	ELSE WRITELN('<!-- License ', s, ' absent. -->'+crlf);

    check_license := chk;
  END;
END;


FUNCTION LENGTH_ASCIZ(val: PACKED ARRAY [l2..u2: INTEGER] OF CHAR;
		      maxlen: [TRUNCATE] INTEGER): INTEGER;
VAR
  i: INTEGER;
BEGIN
  i := INDEX(val, CHR(0)) - 1;
  
  IF (i < 0) THEN i := LENGTH(val);
  
  IF PRESENT(maxlen) AND_THEN (i > maxlen) THEN i := maxlen;
  
  length_asciz := i;
END;


FUNCTION HTML_ASCIZ(pre:    STRING;
		    val:    PACKED ARRAY [l2..u2: INTEGER] OF CHAR;
		    post:   STRING;
		    maxlen: [TRUNCATE] INTEGER): INTEGER;
VAR
  i: INTEGER;
BEGIN
  IF PRESENT(maxlen)
    THEN i := length_asciz(val, maxlen)
    ELSE i := length_asciz(val);
    
  WRITELN(pre, val:i, post);
  
  HTML_asciz := i;
END;


PROCEDURE HTML_DATETIME(pre: STRING; val: $uquad; post:STRING);
VAR
  s: VARYING [80] OF CHAR;
BEGIN
  LIB$FORMAT_DATE_TIME(date_string := %DESCR s, date := val);
  IF (debug>0) THEN WRITELN( pre, s, post);
END;


FUNCTION VERIFY_FIRST_AND_LAST(VAR rab: rab$type
			      ;VAR index_page: INTEGER
			      ;VAR font_page: INTEGER
			      ):UNSIGNED;
VAR
  p1	     : [unsafe]^br_first_record;
  pk	     : [unsafe]^br_key_head;
  pl	     : [unsafe]^br_last_record;
  i	     : INTEGER;
  sts	     : UNSIGNED;
  licensed   : BOOLEAN VALUE FALSE;
  end_offset : UNSIGNED;
BEGIN
  sts := $GET(rab);		(* first record *)
  
  IF ODD(sts) THEN BEGIN
    p1 := rab.rab$l_rbf;
    
    WRITELN('<HTML><HEAD>');
    HTML_ASCIZ('<TITLE>',p1^.title_asciz,'</TITLE>');
    WRITELN('</HEAD><BODY>',crlf);
    WRITELN('<!-- Gateway: '+htbook_version+' -->'+crlf);

    IF (debug>0) THEN WITH p1^ DO BEGIN
      HTML_ASCIZ('<!-- Author: ', author_asciz,' -->'+crlf);
      HTML_ASCIZ('<!-- License: ', product_asciz,' -->'+crlf);
      licensed := CHECK_LICENSE(author_asciz, product_asciz);
      HTML_DATETIME('<!-- Date: ', credate,' -->'+crlf);
      HTML_ASCIZ('<!-- Id: ', id_asciz,' -->'+crlf);
    END;

    i := p1^.len - SIZE(br_first_record);
    pk := p1::INTEGER + SIZE(br_first_record);

    WHILE (pk^.sublen > 0) AND_THEN (i >= pk^.sublen) DO BEGIN

      WITH pk^ DO CASE subtype OF

	sb_license   : BEGIN
	  HTML_ASCIZ('<!-- Alternate license: ', alt_license,' -->',
		     sublen - SIZE(sublen) - SIZE(subtype));

	  IF NOT licensed THEN
	    licensed := CHECK_LICENSE(p1^.author_asciz,
				      alt_license,
				      sublen - SIZE(sublen) - SIZE(subtype));

	END;

	sb_tblhdr    : BEGIN
	  nkeys	    := nkeys + 1;
	  keykind[nkeys]  := keyid;
	  keycount[nkeys] := count;
	  keypart[nkeys]  := part;
	  keyname[nkeys]  := SUBSTR(title_asciz, 1,
				    HTML_asciz('<!-- Keytable: ',
				    title_asciz,' -->',
				    title_b_len));

	  IF (debug>0) THEN
	    WRITELN('  <!-- Type, count, page: ', keyid:1, ', ',
		    count:1, ', ', part:1, ' -->'+crlf);

	END;

	OTHERWISE
	  IF (debug>0)
	    THEN WRITELN('<!-- Unknown subtype ', subtype:1, ' -->'+crlf)
      END;		    

      i := i - pk^.sublen;
      pk := pk::INTEGER + pk^.sublen;
    END;

    IF (i <> 0) AND (debug>0)
      THEN WRITELN('<STRONG>%HTBOOK-E-BADREC1, record 1 incomplete: ', i:1,'</STRONG>');
      
    IF (p1^.rtype <> part_first)
      THEN WRITELN('<STRONG>%HTBOOK-E-BADREC1, record 1 has wrong type: ',
		   p1^.rtype:1,'</STRONG>')
      ELSE
    IF (rab.rab$w_rsz < SIZE(br_first_record))
    OR_ELSE (rab.rab$w_rsz < p1^.len)
      THEN WRITELN('<STRONG>%HTBOOK-E-BADSIZE1, record 1 size: ',
		   rab.rab$w_rsz:1,' not ',p1^.len:1,'</STRONG>');

    index_page := p1^.index_page;
    font_page := p1^.font_page;

    NEW(parts, p1^.partcount);
    NEW(sects, p1^.sectioncount);
    rab.rab$w_rfa := p1^.lastptr;
    rab.rab$b_rac := rab$c_rfa;
    sts := $get(rab);		 (* get last record *)		 

    IF ODD(sts) THEN BEGIN
      pl := rab.rab$l_rbf;
      
      IF (pl^.rtype <> last)
        THEN WRITELN('<STRONG>%HTBOOK-E-BADRECTYP, "last" record has wrong type: ',pl^.rtype:1,'</STRONG>')
      ELSE BEGIN
	end_offset := (BYTE_OFFSET(br_last_record, part) + parts^.n * SIZE(br_ptr));
	IF (end_offset	<> rab.rab$w_rsz)
	  THEN WRITELN('<STRONG>%HTBOOK-E-BADLAST, "last" content length ',end_offset:1,
	  ' does not match record length ',rab.rab$w_rsz:1,'</STRONG>');
      END;
      
      FOR i := 0 TO parts^.n - 2 DO parts^[i] := pl^.part[i];
      
      rab.rab$b_rac := rab$c_seq; (* see if it really is last *)
      sts := $find(rab);
      
      IF (sts = rms$_eof)
        THEN sts := rms$_suc
	ELSE WRITELN('<STRONG>%HTBOOK-W-NOTLAST, "last" record is not last in file.</STRONG>');
      
    END;
    
  END;
  
  verify_first_and_last := sts;
END;


PROCEDURE DISPLAY_XL(pu: pun; i: INTEGER);
BEGIN
  
  IF (i > 0) AND_THEN (debug>0) THEN BEGIN
    WRITE(i:1, ' p (long) = ', pu^: 1);
    
    WHILE (i > 1) DO BEGIN
      i := i - 1;
      pu := pu::INTEGER + 4;
      WRITE(', ', pu^: 1);
    END;
    
    WRITELN(' -->'+crlf);

  END;
END;


PROCEDURE DISPLAY_XW(pu: puw;
		     i: INTEGER);
    BEGIN
    IF (i > 0) AND_THEN (debug>0) THEN
	BEGIN
	WRITE(i:1, 'p (word) = ', pu^: 1);

	WHILE (i > 1) DO
	    BEGIN
	    i := i - 1;
	    pu := pu::INTEGER + 2;
	    WRITE(', ', pu^: 1);
	    END;

	    WRITELN(' -->'+crlf);
	END;
    END;


PROCEDURE DISPLAY_XB(pu: pub;
		     i: INTEGER);
BEGIN
  IF (i > 0) AND_THEN (debug>0) THEN
  BEGIN
    WRITE(i:1, 'p (byte) = ', pu^: 1);
    WHILE (i > 1) DO BEGIN
      i := i - 1;
      pu := pu::INTEGER + 1;
      WRITE(', ', pu^: 1);
    END;
    
    WRITELN(' -->'+crlf);
  END;
END;



FUNCTION HIT(s,x,y:INTEGER):hotspotp;
VAR
  h:hotspotp;
BEGIN
  h:=hotspots;
  WHILE (h<>NIL)
  AND_THEN ((s<>h^.s) OR (h^.x>x) OR (h^.x+h^.w<x) OR (h^.y>y) OR (h^.y+h^.h<y))
    DO h:=h^.pnxt;
  RETURN h;
END;


(* add a string to output; checking for positioning, hotspots, etc *)  
PROCEDURE HTMLSTR( hor,ver:INTEGER; s:STRING );
VAR
  h	  : hotspotp;
  linkstr : STRING(254);
  i	  : INTEGER;
  bigskip : BOOLEAN;
BEGIN
  h:=HIT(cursect,hor,ver);
  IF h<>NIL THEN BEGIN
    WRITEV(linkstr,'<A HREF="', book_url,'/',sects^[h^.target].partno:1,
	  '#', h^.target:1, '"',
	  ' X=', h^.x:1, (* just for debug *)
	  ' Y=', h^.y:1, (* -"- *)
	  ' W=', h^.w:1, (* -"- *)
	  ' H=', h^.h:1, (* -"- *)
	  '>');
    curstr:=curstr+linkstr;
    (* IF we disposed the hotspot now, future searches would be faster... *)
  END;

  bigskip:=ver>(cury+100);
  IF bigskip OR ((ver>cury) AND NOT hyphen) THEN BEGIN	(* going down: *)
    WRITE(curstr);
    curstr:='';
    IF debug>1
      THEN WRITE('<!-- vskip ',ver-cury:1,' -->'+crlf);
    IF bigskip THEN BEGIN
      WRITE('<P>');  (* prob. empty line (1 line@14p=89)*)
      curx:=0;
    END;
    WRITELN(crlf); (* so <PRE> sections will look better *)
  END;
  hyphen:=false;

  (* add to string. leftstep? upstep? *)
  curstr:=curstr+PAD('',' ',MAX(0,ROUND((hor-curx) / scale)));

  (* escape the HTML syntacticals as HTML entities *)
  CASE fonts[curfont].encoding OF
    latin1:FOR i:=1 TO LENGTH(s) DO
	     CASE s[i] OF 
	       '<' : curstr:=curstr+'&lt;';
	       '>' : curstr:=curstr+'&gt;';
	       '&' : curstr:=curstr+'&amp;';
	       '-' : IF (i<>LENGTH(s)) OR_ELSE (i<2) OR_ELSE NOT (s[i-1] IN letters)
		       THEN curstr:=curstr+'-'
		       ELSE hyphen:=TRUE;     (* Just a hyphen. Kill it. *)
	       OTHERWISE curstr:=curstr+s[i];
	     END;
    mathx:FOR i:=1 TO LENGTH(s) DO (* this is a first cut - more to follow *)
	     CASE s[i] OF 
	       'n' : curstr:=curstr+'{';
	       'o' : curstr:=curstr+'}';
	       OTHERWISE curstr:=curstr+s[i];
	     END;
    symbol:FOR i:=1 TO LENGTH(s) DO (* -"- *)
	     CASE s[i] OF 
	       '' : BEGIN    (* Bullet. This is probably a list element :) *)	
		 IF (hor<100) AND NOT bigskip THEN curstr:=curstr+'<BR>';
		 curstr:=curstr+'';
	       END;
	       OTHERWISE curstr:=curstr+s[i];
	     END;
    OTHERWISE curstr:=curstr+s;
  END;	
  cury := ver;
  curx := hor;

  IF h<>NIL THEN curstr:=curstr+'</A>';

END;


PROCEDURE POST_CONTENTS;
BEGIN
  IF (post.length>0) OR (curstr.length>0) THEN BEGIN (* something left to flush? *)
    WRITELN(curstr,post,crlf);			      (* crlf?? *)
    curstr:='';
    post:='';
  END;
  curfont:=-1;
  hyphen:=FALSE;
END; { POST_CONTENTS }



PROCEDURE SETFONT(hor,fontno : INTEGER);
CONST fudge = 100;
BEGIN
  IF fontno<>curfont THEN WITH fonts[fontno] DO BEGIN
    POST_CONTENTS;		(* this is chickening out, but for now... *)
    IF (name="Helvetica") AND in_bold AND (hor < 100) THEN BEGIN (* title? *)
      IF charsize>=180
	THEN BEGIN WRITE('<H1>'); post:='</H1>'+post; END
	ELSE
      IF charsize>=140
	THEN BEGIN WRITE('<H2>'); post:='</H2>'+post; END
	ELSE BEGIN WRITE('<H3>'); post:='</H3>'+post; END;
    END ELSE BEGIN
      IF in_bold    THEN BEGIN WRITE('<B>'); post:='</B>'+post; END;
      IF in_italics THEN BEGIN WRITE('<I>'); post:='</I>'+post; END;
      IF underline  THEN BEGIN WRITE('<U>'); post:='</U>'+post; END;
      IF name="Courier" THEN BEGIN
	IF hor<fudge
	  THEN BEGIN WRITE('<PRE>'); post:='</PRE>'+post; END
	  ELSE BEGIN WRITE('<TT>'); post:='</TT>'+post; END;
      END;
      IF encoding<>latin1 THEN BEGIN WRITE('<!--WEIRD-->'); post:='<!--/WEIRD-->'+post; END;
    END;
    curfont:=fontno;
    scale := MAX(fonts[fontno].charsize*average_charwidth,min_scale);
  END;
END;


PROCEDURE PROCESS_TEXTREC(tp : [unsafe] br_txt_rec_p	(* start here *)
			 ;i  : INTEGER  	        (* this many bytes *)
			 ;ignore_plain : BOOLEAN:=FALSE
		         );
TYPE
  str = PACKED ARRAY [1..255] OF CHAR;
VAR
  pa	     : [unsafe]^ascic;
  pb	     : pub;
  longp	     : pun;
  chars_left : INTEGER;
  s	     : VARYING [1022] OF CHAR;
  tx	     : INTEGER VALUE 0;
BEGIN
  WHILE (i > 0) DO BEGIN
    WITH tp^ DO CASE subsectyp OF
      1:BEGIN			(* rules and other black-filled areas *)
	IF (debug>1)
	  THEN WRITELN( '<!-- Subsect type ',subsectyp:1,': Fill x,y = ', hor:1,',',ver:1,
		      '; w,h = ', fontno:1, ',',y:1, ' -->'+crlf);
	curstr:=curstr+'<HR>';
	i  := i - tp^.len;
	tp := tp::INTEGER + tp^.len; (* step to next subsect *)
      END; 
      
      2,3:BEGIN	(* text *)
	    IF (debug>1) THEN BEGIN
	        WRITELN('	<!-- Subsect type ',subsectyp:1,
			  ', length ', len:1, '; H,V = ', hor:1, ',', ver:1,
			  '; font = ', fontno: 1,'(',fonts[fontno].charsize DIV 10:1,
			  ' p ',fonts[fontno].name,
			  '); ?,?=',x:1, ',', y:1,' -->'+crlf);
	    END;
	    pa := IADDRESS(txt);
	    chars_left := len - BYTE_OFFSET(tp^, txt);
	    
	    SETFONT(hor,fontno);(* Font change: new rendition selection: *)

	    IF NOT ignore_plain THEN BEGIN
	      tx := hor;
	      s:='';
	      WHILE (chars_left > 0) DO BEGIN
		s := s + SUBSTR(pa^.txt, 1, pa^.len);
		tx := tx + ROUND(pa^.len * scale);
		chars_left := chars_left - 1 - pa^.len - 1;
		pb := pa::INTEGER + 1 + pa^.len; (* point to xpad before next word *)

		IF (chars_left >= 0) THEN BEGIN
		  tx := tx + pb^;

		  IF (debug>3)
		      THEN s:=s+''+DEC(pb^,2)+''
		      ELSE s:=s+PAD('',' ',ROUND(pb^ / scale));
		END;
		pa := pb::INTEGER + 1; (* point to next word *)
	      END;
	      
	      HTMLSTR(hor,ver,s);
	    END;
	    i := i - tp^.len;
	    tp := tp::INTEGER + tp^.len;	 (* step to next subsect*)
	  END;

      OTHERWISE	(* simply zero-terminated text followed by a LONG pointer *)
      
	chars_left := LENGTH_ASCIZ(tp^::str, i) + 1;
	
	IF (chars_left < i) THEN BEGIN
	  longp := tp::INTEGER + chars_left;
	  
	  IF (debug>1) THEN
	    WRITE('  <!-- Target ',longp^, '-->' );
	  
	  WRITELN('<P><A HREF="',
		  book_url,'/',sects^[longp^].partno:1,'#',longp^:1,
		  '">',tp^::str:chars_left-1,'</A>');
	END ELSE WRITELN(tp^::str:chars_left-1);
	
	i := 0;
    END;(*case*)
  END;
(*  POST_CONTENTS;*)
END;


PROCEDURE PROCESS_SUBPART_IXTXT(pk: pbr_key_head);
BEGIN
  IF (debug>0) THEN BEGIN
    WRITE( '  <!-- Ixtxt subpart; width ',pk^.ixhdr[1]:1,' height ',pk^.ixhdr[2]:1, '; ');
    display_xw(IADDRESS(pk^.ixhdr), 7);
  END;
  cursect:=0;
  PROCESS_TEXTREC(IADDRESS(pk^.ixtxt), pk^.sublen-BYTE_OFFSET(pk^,ixtxt), ignore_plain:=TRUE);
END;



PROCEDURE PROCESS_SUBPART_BODYTXT(pk: pbr_key_head);
BEGIN
  IF (debug>0) THEN BEGIN
    WRITE( '  <!-- Body txt subpart ',pk^.bodhdr[3]:1,' height ',pk^.bodhdr[8]:1, '; ');
      display_xl(IADDRESS(pk^.bodhdr), 9);
  END;
  (* Make a HTML anchor.*)
  CASE html OF 	
    1,2:WRITELN('<P><A NAME="',pk^.bodhdr[3]:1,'">');
    3:  WRITELN('<P ID="',pk^.bodhdr[3]:1,'">'); (* HTML+ *)
  END;
  cursect := pk^.bodhdr[3];
  sects^[cursect].rfa := rab.rab$w_rfa;	(* remember exact position *)
  sects^[cursect].offset := pk::INTEGER - rab.rab$l_rbf::INTEGER;
  
  PROCESS_TEXTREC(IADDRESS(pk^.bodtxt), pk^.sublen - BYTE_OFFSET(pk^,bodtxt));
  POST_CONTENTS;
  IF html <3 THEN WRITELN('</A>');		(* only for HTML- *)
END;



(* remember the hotspots for later *)
PROCEDURE PROCESS_SUBPART_HOTSPOT(pk: pbr_key_head);
VAR
  hs, hs1: hotspotp;
BEGIN
  WITH pk^ DO BEGIN
    IF (debug>0)
      THEN WRITELN( '<!-- [   Hotspot: ', hot_unk1:1, ',',
		   hot_unk2:1, '; sect=', hot_sect:1, '; x,y=',
		   horpos:1, ',', verpos:1, '; w,h=', len:1, ',',
		   wid:1, '; target=', hot_target:1, '; ', hot_unk3:1, ' (',
		   pk^.sublen-4-BYTE_OFFSET(br_key_head, hot_unk3):1,' bytes) ] -->'+crlf);

    NEW(hs);
    WITH hs^ DO BEGIN(*	chars := ROUND(len/36);*)
      pnxt := NIL;
      pprv := NIL;
      x:=horpos;
      y:=verpos;
      w:=len;
      h:=wid;
      s:=hot_sect;
      target:=hot_target;
    END;

    IF (hotspots=NIL) THEN BEGIN (* link it first*)
      hotspots := hs;
      curhotspot := hs;
    END
    ELSE
    BEGIN			(* sort it into list on ascending y *)
      hs1 := curhotspot;
      WHILE (hs1<>NIL) AND_THEN (hs1^.y>hs^.y)
        DO hs1 := hs1^.pprv;
      IF (hs1<>NIL) THEN
        WHILE (hs1<>NIL) AND_THEN (hs1^.y=hs^.y) DO hs1 := hs1^.pprv;
      IF (hs1=NIL) THEN BEGIN	(* least y, so put it first *)
	hs^.pnxt := hotspots;
	hotspots^.pprv := hs;
	hotspots := hs;
      END
      ELSE
      BEGIN
	hs^.pnxt := hs1^.pnxt;
	hs^.pprv := hs1;
	hs1^.pnxt := hs;
	IF (hs^.pnxt=NIL)
	  THEN curhotspot := hs
	  ELSE hs^.pnxt^.pprv := hs;
      END;
    END;
  END;(* WITH pk^ *)
END;{}


(* remember the grayspots for later *)
PROCEDURE PROCESS_SUBPART_GRAYSPOT(pk: pbr_key_head);
VAR
  hs, hs1: hotspotp;
  i: INTEGER;
BEGIN
  WITH pk^ DO BEGIN
    IF (debug>0)
      THEN WRITELN('<!-- [ Grayspot : ', ext_unk1:1, ',',
		   ext_unk2:1, '; sect=', ext_sect:1, '; x,y=',
		   ext_x:1, ',', ext_y:1, '; w,h=', ext_w:1, ',',
		   ext_y:1, '; ',ext_unk3:1, ', ', ext_unk4:1, ' (',
		   pk^.sublen-4-BYTE_OFFSET(br_key_head, hot_unk3):1,
		   ' bytes) ] -->'+crlf);

(*    NEW(hs);
    WITH hs^ DO BEGIN
      pnxt := NIL;
      pprv := NIL;
      x:=horpos;
      y:=verpos;
      w:=len;
      h:=wid;
      s:=ext_sect;		(* section, really * )
    END;

    IF (hotspots=NIL) THEN BEGIN (* link it first* )
      hotspots := hs;
      curhotspot := hs;
    END
    ELSE
    BEGIN			(* sort it into list on ascending y * )
      hs1 := curhotspot;
      WHILE (hs1<>NIL) AND_THEN (hs1^.y>hs^.y)
        DO hs1 := hs1^.pprv;
      IF (hs1<>NIL) THEN
        WHILE (hs1<>NIL) AND_THEN (hs1^.y=hs^.y) DO hs1 := hs1^.pprv;
      IF (hs1=NIL) THEN BEGIN	(* least y, so put it first * )
	hs^.pnxt := hotspots;
	hotspots^.pprv := hs;
	hotspots := hs;
      END
      ELSE
      BEGIN
	hs^.pnxt := hs1^.pnxt;
	hs^.pprv := hs1;
	hs1^.pnxt := hs;
	IF (hs^.pnxt=NIL)
	  THEN curhotspot := hs
	  ELSE hs^.pnxt^.pprv := hs;
      END;
    END;
*)
  END;(* WITH pk^ *)
END; { PROCESS_SUBPART_GRAYSPOT }
  

PROCEDURE PROCESS_SUBPART_UNKNOWN(pk: [unsafe] pbr_key_head);
VAR
  i: INTEGER;
  pn: pun;
BEGIN
  WITH rab, pk^
    DO IF (debug>0)
      THEN WRITELN('<!-- Unknown subpart ', subtype, ' (', sublen:1,
		   '), RFA = %x', HEX(rab$w_rfa[1],4,4), ',', HEX(rab$w_rfa[2],4,4),
		   ',', HEX(rab$w_rfa[3],4,4), '; offset = %X',
		   HEX(pk::UNSIGNED-rab$l_rbf,4,4),' -->'+crlf);

  i := pk^.sublen - subpart_minlen;
  IF (i>120) THEN i := 120;
  pn := pk::INTEGER + subpart_minlen; (* skip header *)
  WHILE (i>=40) DO BEGIN	(* 10 longs at a time *)
    WRITE('<!-- ');
    display_xl(pn, 10);
    i := i - 40;
    pn := pn::INTEGER + 40;
  END;
  IF (i>=4) THEN BEGIN
    WRITE('<!-- ');
    display_xl(pn, i DIV 4);
    pn := pn::INTEGER + 4*(i DIV 4);
    i := i MOD 4;
  END;
  IF (i>0) THEN BEGIN WRITE('<!-- '); DISPLAY_XB(pn, i); END;
END;


PROCEDURE OUTPUT_XBM(pk	: pbr_key_head; as:STRING);
VAR
  r,c,i:INTEGER;
  picf:TEXT;
BEGIN
  WITH pk^ DO BEGIN
    (* save the picture to a file; this is almost too easy in XBM format, *)
    (* and wastes bandwidth like hell, but it works... *)
    OPEN(picf,'BOOK_FIGURES:'+as,
	 history:=new,record_type:=stream_lf);
    REWRITE(picf);
    WRITELN(picf,'#define ',book_id,'_width ',fig_columns:1);
    WRITELN(picf,'#define ',book_id,'_height ',fig_rows:1);
    WRITELN(picf,'static char ',book_id,'_bits[] = { ');
    FOR i:=1 TO fig_picbytes-8-1 DO BEGIN
      WRITE(picf,'0x',HEX(fig_pixels[i],2),',');
      IF (i MOD 12) = 0 THEN WRITELN(picf);
    END;
    WRITELN(picf,'0x',HEX(fig_pixels[fig_picbytes-8],2),'};');
    CLOSE(picf);
  END;
END; { OUTPUT_XBM }
  

PROCEDURE OUTPUT_GIF_VIA_PBM(pk	: pbr_key_head; as:STRING);
CONST
  lf = CHR(10);
  nowild=1;
VAR
  sts	 : INTEGER;
  i,r,c	 : INTEGER;
  picf	 : TEXT;
  ctx	 : UNSIGNED VALUE 0;
  result : VARYING[255] OF CHAR;
BEGIN
  (* first, see if target file already exists; if so, skip this procedure *)
  sts:=LIB$FIND_FILE('BOOK_FIGURES:'+as+'.GIF',result.body,ctx,flags:=nowild);
  LIB$FIND_FILE_END(ctx);
  IF sts=RMS$_FNF				   (* have to create it *)   
    THEN WITH pk^ DO BEGIN
      (* save the picture to a file in PBM format, *)
      OPEN(picf,'BOOK_FIGURES:'+as+'.PBM'
	  ,history:=new
	  ,record_type:=stream_lf
	  ,record_length:=8192	(* should suffice for most pictures... *)
	(*,carriage_control:=none*)
	  );
      REWRITE(picf);
      WRITELN(picf,'P1');
      WRITELN(picf,fig_columns:1,' ',fig_rows:1);

      FOR r:=0 TO fig_rows-1 DO BEGIN
	i:=(fig_columns+7) DIV 8 * r;		(* start index of this row *)
	FOR c:=0 TO fig_columns-1 DO
	  IF fig_pixels[i+(c DIV 8),c MOD 8]
	    THEN WRITE(picf,'1')		(* This is REALLY inefficient... *)
	    ELSE WRITE(picf,'0');
	WRITELN(picf);
      END;
      CLOSE(picf);
      sts:=LIB$SPAWN('@BOOK_FIGURES:CVT_TO_GIF.COM '+as+'.PBM  '+as+'.GIF');
      IF NOT ODD(sts) THEN LIB$STOP(sts);
    END;
END; { OUTPUT_GIF_VIA_PBM }
  

(* This is faster, assures the file exists when the link is sent
 *  (multithreaded browsers, rembember?) and allows other formats
 *  to be added 
 *)

PROCEDURE OUTPUT_GIF_VIA_PBM_MBX(pk : pbr_key_head; name:STRING);
CONST
  lf	 = CHR(10);
  nowild = 1;
VAR
  sts	  : INTEGER;
  sub_sts : [volatile]INTEGER;
  efn	  : INTEGER;
  i,r,c	  : INTEGER;
  ctx	  : UNSIGNED VALUE 0;
  buf	  : STRING(pk^.fig_columns+8);
  mchn	  : $UWORD;
  result  : VARYING[255] OF CHAR;
BEGIN
  (* first, see if target file already exists; if so, skip this procedure *)
  sts:=LIB$FIND_FILE('BOOK_FIGURES:'+name+'.GIF',result.body,ctx,flags:=nowild);
  LIB$FIND_FILE_END(ctx);

  IF sts=RMS$_FNF				   (* have to create it *)   
    THEN WITH pk^ DO BEGIN
      ASSERT($CREMBX(chan:=mchn,		   (* get a mailbox *)		    
		     maxmsg:=256,
	             lognam:='HTBOOK_MBX', flags:=CMB$M_WRITEONLY));
      ASSERT(LIB$GET_EF(efn::UNSIGNED));	   (* get an event flag *)
      ASSERT($CLREF(efn));
      ASSERT(LIB$SPAWN('@BOOK_FIGURES:CVT_TO_GIF.COM '+DEC(efn)+' '+name+'.GIF',
		       output_file:='BOOK_FIGURES:CVT.LOG',
		       flags:=CLI$M_NOWAIT+CLI$M_NOKEYPAD+CLI$M_NOCLISYM,
		       completion_status_address:=sub_sts,
		       byte_integer_event_flag_num:=efn
	               ));
      ASSERT($QIOW(chan:=mchn, func:=IO$_WRITEVBLK, P1:='P1', P2:=2 ));   (* magic # *)
      WRITEV(buf,fig_columns:1,' ',fig_rows:1);
      ASSERT($QIOW(chan:=mchn,func:=IO$_WRITEVBLK,P1:=buf.body,P2:=buf.length));

      FOR r:=0 TO fig_rows-1 DO BEGIN
	buf:='';
	i:=(fig_columns+7) DIV 8 * r;		(* start index of this row *)
	FOR c:=0 TO fig_columns-1 DO BEGIN
	  IF fig_pixels[i+(c DIV 8),c MOD 8]
	    THEN buf:=buf+'1'
	    ELSE buf:=buf+'0';
	  IF LENGTH(buf)>=200
	    THEN BEGIN
	      ASSERT($QIOW(chan:=mchn,func:=IO$_WRITEVBLK,P1:=buf.body,P2:=buf.length));
	      buf:='';
	    END;
	END;
	ASSERT($QIOW(chan:=mchn,func:=IO$_WRITEVBLK,P1:=buf.body,P2:=buf.length));
      END;
      ASSERT($QIOW(chan:=mchn,func:=IO$_WRITEOF+IO$M_READERCHECK));(* maybe IT closed *)
      ASSERT($DASSGN(mchn));
      ASSERT($WAITFR(efn));		 (* now wait for the subprocess to die *)
   (*ASSERT(sub_sts);   (* seems to return CLI-F-BADWINCNT, bad window pointer count!*)
    END;
END; { OUTPUT_GIF_VIA_PBM_MBX }
  

PROCEDURE PROCESS_SUBPART_FIGURE( pk: pbr_key_head );
(*    sb_figure:(fig_unk1, fig_unk2:	 INTEGER;
		 fig_sect,
		 fig_x, fig_y, fig_w, fig_h :INTEGER;
		 fig_unk3, fig_picbytes: INTEGER;
		 fig_primitive, fig_unk5, fig_unk6, fig_unk7:	 $ubyte;
		 fig_columns, fig_rows:	 $uword;
		 fig_pixels:		 PACKED ARRAY [1..65536] OF $ubyte);
*)
VAR
  fn : STRING(80);
  tp : [unsafe]record_p;
BEGIN
  WITH pk^ DO BEGIN
    IF (debug>0) THEN
      WRITE('<!-- Figure, size = ', sublen:1, '; ',
	    fig_unk1:1, ',', fig_unk2:1,
	    '; section=',fig_sect:1,
	    '; x,y=', fig_x:1, ',',fig_y:1,
	    '; w,h=', fig_w:1, ',',fig_h:1,
	    '; ',fig_unk3:1, '; ', fig_picbytes:1, ' bytes; ',
	    HEX(fig_unk5,2),'; ', HEX(fig_unk6,2),'; ', HEX(fig_unk7,2),'; '
	    );

    CASE fig_primitive OF
      %x'4B':BEGIN		(* BITMAP *)
	POST_CONTENTS;
	IF debug>0
	  THEN WRITELN(' bitmap; c,r=', fig_columns:1, ',',fig_rows:1, ' -->'+crlf);
	fn:=book_id+'_'+DEC(fig_sect,8)+'_'+HEX(pk,8);(* probably unique... *)
	OUTPUT_GIF_VIA_PBM_MBX(pk,fn);
	WRITELN('<IMG SRC="/book_figures/',fn,'.GIF">');
      END;
      %x'03':BEGIN			(* graphical TXT, I guess... *)
	IF debug>0 THEN WRITELN(' graphtext; font=',fig_rows:1,' ?=',fig_columns:1,'-->'+crlf);
	tp:=IADDRESS(fig_pixels)+2;
	(*WRITELN(tp^:fig_pixels[1]::$UBYTE); (* asciw?? *)
	SETFONT(fig_x,fig_rows);
	HTMLSTR(fig_x,fig_y,SUBSTR(tp^,1,fig_pixels[1]::$UBYTE));
      END;
      OTHERWISE
	POST_CONTENTS;
        IF debug>0 THEN WRITELN(' unknown primitive:',HEX(fig_primitive,2),'; ',
			        '; ?,?=', fig_columns:1, ',',fig_rows:1,
				' -->'+crlf);
    END;
  END;
END;

PROCEDURE EXTENDPART(VAR pk, len: [unsafe] INTEGER);
VAR
  i: INTEGER;
BEGIN
  i := broken_subpart^.capacity - broken_subpart^.LENGTH;
  IF (i>len) THEN i := len;
  broken_subpart^ := broken_subpart^ + SUBSTR(pk::record_p^, 1, i);
  len := len - i;
  pk := pk + i;
END;


PROCEDURE EXTENDVARPART(VAR pk, len: [unsafe] INTEGER);
VAR
  ps: ^STRING;
  pk1: [unsafe] pbr_key_head;
BEGIN
  EXTENDPART(pk, len);
  pk1 := IADDRESS(broken_subpart^.body);

  IF (LENGTH(broken_subpart^) < pk1^.sublen)
  AND_THEN (len>0) THEN BEGIN
    ps := broken_subpart;
    NEW(broken_subpart, pk1^.sublen);
    broken_subpart^ := ps^;
    DISPOSE(ps);
    extendpart(pk, len);
  END;
END;


PROCEDURE MAKEPART( pk: [unsafe] record_p; cap, len: INTEGER );
BEGIN
  NEW(broken_subpart, cap);
  broken_subpart^ := SUBSTR(pk^, 1, len);
END;


PROCEDURE MAKEVARPART( pk: [unsafe] pbr_key_head; len: INTEGER );
BEGIN
  IF (len<subpart_minlen)
    THEN MAKEPART(pk, subpart_minlen, len)
    ELSE MAKEPART(pk, pk^.sublen, len);
END;


PROCEDURE DISPLAY_CONTENT_SUBPARTS( pk: [unsafe]pbr_key_head; i:INTEGER );
VAR
  ps: ^STRING;
  pk1: [unsafe]pbr_key_head;
BEGIN
  IF (broken_subpart <> NIL) THEN BEGIN
    EXTENDVARPART(pk, i);
    pk1 := IADDRESS(broken_subpart^.body);
    IF (pk1^.sublen<=LENGTH(broken_subpart^)) THEN BEGIN
      ps := broken_subpart;
      broken_subpart := NIL;
      DISPLAY_CONTENT_SUBPARTS(pk1, pk1^.sublen); (* recurse *)
      DISPOSE(ps);
    END;
  END;

  WHILE (i > 0) AND_THEN (pk^.sublen > 0) DO BEGIN
    IF (i < pk^.sublen)
      THEN makevarpart(pk, i)
      ELSE CASE pk^.subtype OF
	
	sb_ixtxt: PROCESS_SUBPART_IXTXT(pk);
	
	sb_bodytext: PROCESS_SUBPART_BODYTXT(pk);
	
	sb_figure: PROCESS_SUBPART_FIGURE(pk);
	
	sb_hotspot:IF debug>0 THEN WRITELN('<!-- [hotspot here] -->'+crlf);
	(* process_subpart_fig_hotspot(pk) *); 
	
	sb_extension:IF debug>0 THEN WRITELN('<!-- [grayspot here] -->'+crlf);
	(* process_subpart_grayspot*)
	
	OTHERWISE
	  process_subpart_unknown(pk);
      END;
      
      i := i - pk^.sublen;
      pk := pk::INTEGER + pk^.sublen;
  END;
END;


(* Do an ordinary text page, with pointers to previous and next *)
PROCEDURE PROCESS_BODYPAGE(p: pun; len: INTEGER; firstrec,lastrec: BOOLEAN);
(*bodypart_hdr = PACKED RECORD
    partno      : INTEGER;
    unknown_one : INTEGER;
    sectcnt	: INTEGER;
    prevpart    : INTEGER;
    nextpart    : INTEGER;
  END;*)
VAR
  hdr: [unsafe]^bodypart_hdr;
  i : INTEGER;
  prevpart, nextpart: INTEGER VALUE ZERO;
BEGIN
  IF firstrec THEN BEGIN	(* interpret and skip header *)
    hdr := p;
    prevpart := hdr^.prevpart;
    nextpart := hdr^.nextpart;
    IF (debug>0) THEN
        WRITELN('<!-- Page ', hdr^.partno:1, ' (',hdr^.unknown_one:1, '); ', hdr^.sectcnt: 1,
			     ' sections ', 
			     '(Previous=', prevpart:1,
			     '; Next=', nextpart:1,') -->'+crlf);

    p := p::INTEGER + SIZE(hdr^);
    len := len - SIZE(hdr^);

    (*  the HTML+ <LINK> will be better*)
    IF prevpart>0
      THEN WRITE('<A HREF="',book_url,'/',prevpart:1,'">Previous page</A> ');
    IF nextpart>0
      THEN WRITE('<A HREF="',book_url,'/',nextpart:1,'">Next page</A> ');
    FOR i:=1 TO nkeys
      DO WRITELN('<A HREF="',book_url,'/',keypart[i]:1,'">',keyname[i],'</A> ');
    WRITE('<HR>');
  END;

  DISPLAY_CONTENT_SUBPARTS(p, len);
  POST_CONTENTS;

  IF lastrec AND NOT firstrec THEN BEGIN(* long page; repeat links *)
    WRITE('<HR>');
    IF prevpart>0
      THEN WRITE('<A HREF="',book_url,'/',prevpart:1,'">Previous page</A> ');
    IF nextpart>0
      THEN WRITE('<A HREF="',book_url,'/',nextpart:1,'">Next page</A> ');
    FOR i:=1 TO nkeys DO
      WRITELN('<A HREF="',book_url,'/',keypart[i]:1,'">',keyname[i],'</A> ');
  END;
END;


PROCEDURE FIND_HOTSPOT_SUBPARTS(pk    : [unsafe] pbr_key_head;
				i     : INTEGER);
VAR
  ps: ^STRING;
  pk1: [unsafe] pbr_key_head;
BEGIN
  IF (broken_subpart <> NIL) THEN BEGIN
    EXTENDVARPART(pk, i);
    pk1 := IADDRESS(broken_subpart^.body);
    IF (pk1^.sublen<=LENGTH(broken_subpart^)) THEN BEGIN
      ps := broken_subpart;
      broken_subpart := NIL;
      FIND_HOTSPOT_SUBPARTS(pk1, pk1^.sublen); (* recurse *)
      DISPOSE(ps);
    END;
  END;
  
  WHILE (i > 0) AND_THEN (pk^.sublen > 0) DO BEGIN
    IF (i < pk^.sublen)
      THEN MAKEVARPART(pk, i)
      ELSE CASE pk^.subtype OF
	sb_hotspot:   PROCESS_SUBPART_HOTSPOT(pk);
	sb_extension: PROCESS_SUBPART_GRAYSPOT(pk);
	OTHERWISE;
      END;
      i := i - pk^.sublen;
      pk := pk::INTEGER + pk^.sublen;
  END;
END;


(* Scan an ordinary text page for hotspots *)
PROCEDURE FIND_PAGE_HOTSPOTS(p: pun; len: INTEGER; cont: BOOLEAN);
TYPE
  p5 = ARRAY [1..5] OF UNSIGNED; (*thispart,1,sectcount,prevpart,nextpart*)
VAR
  pk: [unsafe]^p5;
  i : INTEGER;
BEGIN
  IF NOT cont THEN BEGIN	(* skip over header *)
    pk := p;
    IF (debug>0) THEN
        WRITELN('<!-- Hotpots on page ', pk^[1]:1, ' -->'+crlf);
    p := p::INTEGER + SIZE(p5);
    len := len - SIZE(p5);
  END;
  FIND_HOTSPOT_SUBPARTS(p, len);
END;


PROCEDURE PROCESS_PAGE_INDEX(pk: pbr_key_head; len: INTEGER; cont: BOOLEAN);
BEGIN
  DISPLAY_CONTENT_SUBPARTS(pk, len);
END;


PROCEDURE BUILD_SECTION_MAP(pk: pun; len: INTEGER; cont: BOOLEAN);
VAR
  nxtsct: [STATIC] INTEGER VALUE 0;
  i, l: INTEGER;
  pk1: pun;
BEGIN
    IF NOT cont THEN
	nxtsct := 0
    ELSE IF (broken_subpart<>NIL) THEN BEGIN
      extendpart(pk, len);
      pk1 := IADDRESS(broken_subpart^.body);
      
      IF (debug>2) THEN
	WRITELN('<!-- And Section ', nxtsct:1, ' is on page ', pk1^: 1,' -->'+crlf);
      
      WITH sects^[nxtsct] DO BEGIN
	rfa := ZERO;		(* don't know where exactly it is yet... *)
	offset := 0;
	partno := INT(pk1^);
      END;
      nxtsct := nxtsct + 1;
    END;
  
    l := nxtsct - 1 + (len DIV 4);

    FOR i := nxtsct TO l DO BEGIN
      
      IF (debug>2) THEN
	WRITELN('<!-- Section ', i: 1, ' is on page ', pk^:1, ' -->'+crlf);
      
      WITH sects^[i] DO BEGIN
	rfa := ZERO;		(* don't know where exactly it is yet... *)
	offset := 0;
	partno := INT(pk^);
      END;
      
      pk := pk::INTEGER + 4;
    END;
  
  nxtsct := l + 1;
  i := len MOD 4;
  
  IF (i>0) THEN MAKEPART(pk, 4, i);

END;


PROCEDURE PROCESS_PAGE_FDESC(pk: [unsafe] pbr_fdesc; len: INTEGER; cont: BOOLEAN);
VAR
  j   : [STATIC] INTEGER VALUE 1;
  pk1 : [unsafe] pbr_fdesc;
BEGIN
  IF NOT cont THEN
    j := 1
  ELSE IF (broken_subpart<>NIL) THEN BEGIN
    EXTENDVARPART(pk, len);
    pk1 := IADDRESS(broken_subpart^.body);
    IF (debug>0) THEN WRITE( 'Section ', j: 1);
    
    j := j + 1;
    HTML_ASCIZ('<!-- FDesc = ', pk1^.txt, ' -->', pk1^.len - BYTE_OFFSET(br_fdesc,txt));
  END;
  
  WHILE (len >= subpart_minlen) AND_THEN (len >= pk^.len) DO BEGIN
    IF (pk^.subtype = sb_fdesc) THEN BEGIN
      IF (debug>0) THEN WRITE( 'Section ', j: 1);
      
      j := j + 1;
      HTML_ASCIZ('<!-- FDesc = ', pk^.txt, ' -->', pk^.len - BYTE_OFFSET(br_fdesc,txt));
    END
    ELSE
      WRITELN('<STRONG>%HTBOOK-E-BADRECTYP, subpart: ', pk^.subtype,'</STRONG>');
      
    len := len - pk^.len;
    pk := pk::INTEGER + pk^.len;
  END;
END;


PROCEDURE PROCESS_PAGE_FONT(pk: [unsafe] pbr_font; len: INTEGER; cont: BOOLEAN);
(*
 Font = -author-family-weight-r-normal--*-size-*-*-p-*-std-ver
 author = ADOBE
 family = Helvetica/Courier/New Century Schoolbook/Symbol/Interim DM etc
 weight = Medium/Bold/* etc
 slant = I/O/R (Italic/Oblique/Roman)
 normal = Normal/* etc
 size = 120/140/180/240 etc (decipoints)
 p = P/M/C (Proportional/Monospaced/Character cell)
 std = ISO8859/*/DEC etc
 ver = 1/DECMATH_SYMBOL/DECMATH_EXTENSION/* etc
*)
VAR
  sep : [STATIC] PACKED ARRAY [1..1] OF CHAR VALUE '-';
  sts : UNSIGNED;
  author, weight, slant,
  normal, null, star1,
  ssize, star2, star3,
  p, star4, std, xtra: VARYING [10] OF CHAR VALUE '';
  family, fontname, ver: VARYING [255] OF CHAR VALUE '';
  ps: ^STRING;
  pk1: [unsafe] pbr_font;
  
BEGIN
  IF cont AND_THEN (broken_subpart <> NIL) THEN BEGIN
    EXTENDVARPART(pk, len);
    ps := broken_subpart;
    broken_subpart := NIL;
    pk1 := ps;
    PROCESS_PAGE_FONT(pk1, pk1^.len, cont);
    DISPOSE(ps);
  END;
  
  WHILE (len > subpart_minlen) AND_THEN (len >= pk^.len) DO BEGIN
    
    IF (pk^.subtype = sb_font) THEN BEGIN
      maxfont := maxfont + 1;
      fonts[maxfont].render := 0;
      fontname := SUBSTR(pk^.txt, 1, length_asciz(pk^.txt,
			 pk^.len - BYTE_OFFSET(br_font, txt)));
      sts := str$element(%DESCR author, 1, sep, %DESCR fontname);
      
      IF ODD(sts)
        THEN sts := str$element(%DESCR family, 2, sep, %DESCR fontname);
	
      IF ODD(sts)
        THEN sts := str$element(%DESCR weight, 3, sep, %DESCR fontname);

	IF ODD(sts) THEN
	    sts := str$element(%DESCR slant, 4, sep, %DESCR fontname);

	IF ODD(sts) THEN
	    sts := str$element(%DESCR normal, 5, sep, %DESCR fontname);

	IF ODD(sts) THEN
	    sts := str$element(%DESCR null, 6, sep, %DESCR fontname);

	IF ODD(sts) THEN
	    sts := str$element(%DESCR star1, 7, sep, %DESCR fontname);

	IF ODD(sts) THEN
	    sts := str$element(%DESCR ssize, 8, sep, %DESCR fontname);

	IF ODD(sts) THEN
	    sts := str$element(%DESCR star2, 9, sep, %DESCR fontname);

	IF ODD(sts) THEN
	    sts := str$element(%DESCR star3, 10, sep, %DESCR fontname);

	IF ODD(sts) THEN
	    sts := str$element(%DESCR p, 11, sep, %DESCR fontname);

	IF ODD(sts) THEN
	    sts := str$element(%DESCR star4, 12, sep, %DESCR fontname);

	IF ODD(sts) THEN
	    sts := str$element(%DESCR std, 13, sep, %DESCR fontname);
				(* ought to uppercase this *)
	IF ODD(sts) THEN
	    sts := str$element(%DESCR ver, 14, sep, %DESCR fontname);
				(* ought to uppercase this *)

	IF ODD(sts) THEN
	    sts := str$element(%DESCR xtra, 15, sep, %DESCR fontname);

	WITH fonts[maxfont] DO BEGIN
	  READV(ssize, charsize, xtra, error := CONTINUE);
	  in_bold:= (weight = 'Bold');
	  in_italics:= (slant = 'I') OR (slant = 'O');
	  encoding:=other;
	  IF (std='ISO8859') AND (ver='1') THEN encoding:=latin1;
	  IF (family='Symbol') THEN encoding:=symbol;
	  IF (std='DEC')     AND (ver='DECMATH_EXTENSION') THEN encoding:=mathx;
	  IF (std='DEC')     AND (ver='DECMATH_ITALIC') THEN encoding:=mathi;
	  IF (std='DEC')     AND (ver='DECMATH_SYMBOL') THEN encoding:=maths;
	  name:=family;
	END;

	IF (debug>0)
	  THEN WRITELN('<!-- Font ', pk^.fontno:1, ' = ', fontname, ' -->'+crlf);
      END
    ELSE
      WRITELN('<STRONG>%HTBOOK-E-BADRECTYP, subpart: ', pk^.subtype,'</STRONG>');

    len := len - pk^.len;
    pk := pk::INTEGER + pk^.len;
  END;
  
  IF (len>0) THEN MAKEVARPART(pk, len);
END;


PROCEDURE PROCESS_PAGE_AZ32(pk: [unsafe] pbr_a32; len: INTEGER; cont: BOOLEAN);
VAR
  i, l: INTEGER;
  nxtsct: [STATIC] INTEGER VALUE 1;
BEGIN
  IF cont THEN BEGIN
    IF (broken_subpart<>NIL) THEN BEGIN
      EXTENDPART(pk, len);
      IF (debug>0) THEN WRITE( '<!-- Section ', nxtsct:1, ' -->'+crlf);
      
      HTML_ASCIZ('<!-- DDesc = ', broken_subpart^.body, ' -->'+crlf);
      DISPOSE(broken_subpart);
      broken_subpart := NIL;
      nxtsct := nxtsct + 1;
    END;
  END
  ELSE
    nxtsct := 1;

  l := nxtsct - 1 + (len DIV SIZE(br_a32));

  FOR i := nxtsct TO l DO BEGIN
    IF (debug>0) THEN WRITE( '<!-- AZ32 Section ', i:1, ' -->'+crlf);
    html_asciz('<!-- DDesc = ', pk^, ' -->'+crlf);
    pk := pk::INTEGER + SIZE(br_a32);
  END;

  i := len MOD SIZE(br_a32);
  IF (i>0) THEN MAKEPART(pk, SIZE(br_a32), i);
  nxtsct := l + 1;
END;


PROCEDURE SCRAP_HOTSPOTS;
BEGIN
  WHILE hotspots<>NIL DO BEGIN
    curhotspot:=hotspots^.pnxt;
    DISPOSE(hotspots);
    hotspots:=curhotspot;
  END;
END;


PROCEDURE PROCESS_PAGE(part: INTEGER
		      ;VAR rab: rab$type
		      ;kind: [TRUNCATE] rectyp
		      );
VAR
  sts: UNSIGNED;
  i, partlen: INTEGER;
  pr: [unsafe] br_record_p;
  pk: [unsafe] pbr_key_head;
  rtyp: rectyp_w VALUE type_unknown;
  cont,done: BOOLEAN;
  pass:INTEGER;
  
BEGIN
  SCRAP_HOTSPOTS;
  pass:=1; done:=TRUE;		(* if we are lucky... *)
  REPEAT
    cury:=0;
    curx:=0;
    curfont:=-1;		(* force immediate 'new font'-processing *)
    cont:=FALSE;		(* this is the first RMS record of the part *)
    rab.rab$b_rac := rab$c_rfa;
    rab.rab$w_rfa := parts^[part].rfa;
    partlen := parts^[part].siz - 16;

    IF (debug>0) THEN WRITELN( '<!-- %%%%% Part ', part:1, ' pass ',pass:1,' %%%%% -->'+crlf);

    REPEAT
      sts := $get(rab);		(* get next record of this page *)
      rab.rab$b_rac := rab$c_seq; (* switch to sequential reading *)
      partlen := partlen + 16 - rab.rab$w_rsz;

      IF NOT ODD(sts)
	THEN LIB$SIGNAL(sts, INT(rab.rab$l_stv))
	ELSE BEGIN
	  pr := rab.rab$l_rbf;

	  IF (rtyp = type_unknown)
	    THEN
	  IF PRESENT(kind)	(* force record type *)
	    THEN rtyp := kind
	    ELSE rtyp := pr^.rtype;

	  IF (partlen < 0) THEN
	     WRITELN('<STRONG>%HTBOOK-E-BADPART, #',part:1,
		     '; RFA ',parts^[part].rfa[1],
		     ',',parts^[part].rfa[2],
		     ',',parts^[part].rfa[3],
		     '; size ',parts^[part].siz:1,
		     '</STRONG>')
	  ELSE IF (pr^.len <> rab.rab$w_rsz) THEN
	     WRITELN('<STRONG>%HTBOOK-E-BADSIZE, #',part:1,
		     '; record size ',rab.rab$w_rsz:1,
		     ' <> ',pr^.len:1,
		     '</STRONG>')
	  ELSE IF (rtyp <> pr^.rtype) AND_THEN
	    (pr^.rtype <> continue_mid) AND_THEN
	    (pr^.rtype <> continue_end) THEN
	     WRITELN('<STRONG>%HTBOOK-E-PARTTYPMIS, #',part:1,
		     '; rtyp ',rtyp:1,
		     'does not match ',pr^.rtype,
		     '</STRONG>')
	  ELSE BEGIN
	    i := pr^.len -	SIZE(br_record); (* step over record header *)
	    pk := pr::INTEGER + SIZE(br_record);
	    IF (partlen>0)
	      THEN i := i - SIZE(br_ptr); (* all but last record ends with fwd ptr *)

	    CASE rtyp OF

	      part_body:IF pass=1
		THEN BEGIN FIND_PAGE_HOTSPOTS(pk, i, cont);done:=FALSE;END
		ELSE PROCESS_BODYPAGE(pk, i, NOT cont, (partlen=0));

	      part_index: PROCESS_PAGE_INDEX(pk, i, cont);

	      part_secmap: BUILD_SECTION_MAP(pk, i, cont);

	      type_f_desc: PROCESS_PAGE_FDESC(pk, i, cont);

	      part_font: PROCESS_PAGE_FONT(pk, i, cont);

	      asciz_32: PROCESS_PAGE_AZ32(pk, i, cont);

	      OTHERWISE
		  WRITELN('%HTBOOK-W-UNKPART, ', rtyp:1 );
	    END;
	  END;
	  cont := TRUE;		(* for next iteration *)
        END;

    UNTIL (partlen <= 0) OR_ELSE NOT ODD(sts);
    pass:=pass+1;
  UNTIL done OR_ELSE (pass>2) OR_ELSE NOT ODD(sts);
END;




FUNCTION OPEN_BOOK(filename: VARYING [u] OF CHAR;
		   default_file: VARYING [w] OF CHAR):INTEGER;
VAR
  sts		: UNSIGNED;
  index_page	: INTEGER;
  font_page	: INTEGER;
BEGIN
  fab.fab$l_fna := IADDRESS(filename.body);
  fab.fab$b_fns := LENGTH(filename);
  
  fab.fab$l_dna := IADDRESS(default_file.body);
  fab.fab$b_dns := LENGTH(default_file);
  
  (*	Set the RMS network block count to a big value (allows up to 65536) *)
  net_block_count := 127;		(* Pointed to by $XABITM *)
  
  sts := $OPEN(fab);

  IF NOT ODD(sts)
    THEN LIB$SIGNAL(sts, INT(rab.rab$l_stv))
    ELSE BEGIN			(* book file opened successfully *)
      IF (xab.xab$w_lrl = 0)
        THEN IF (xab.xab$w_mrz = 0)
	       THEN rab.rab$w_usz := %xFFFF
	       ELSE rab.rab$w_usz := xab.xab$w_mrz
        ELSE rab.rab$w_usz := xab.xab$w_lrl;
      sts := LIB$GET_VM(number_of_bytes := rab.rab$w_usz,    (* Allocate record buffer *)
			base_address := rab.rab$l_ubf);

      IF NOT ODD(sts)
	THEN LIB$SIGNAL(sts)
	ELSE BEGIN
	  sts := $CONNECT(rab);

	  IF ODD(sts) THEN BEGIN
	    (* build sect map; show title; get pages of font and index *)
	    sts := VERIFY_FIRST_AND_LAST(rab, index_page, font_page);

	    IF ODD(sts) THEN BEGIN
	      maxfont := - 1;
	      PROCESS_PAGE(index_page, rab, part_secmap); (* remember index *)
	      PROCESS_PAGE(font_page, rab, part_font); (* remember font data *)
	    END;
	  END;
	END;
    END;
  IF NOT ODD(sts) THEN LIB$SIGNAL(sts, INT(rab.rab$l_stv));
  OPEN_BOOK:=sts;
END;


PROCEDURE CLOSE_BOOK;
BEGIN
  WRITELN('</BODY></HTML>');
(*  DISPOSE(sects);
  DISPOSE(parts);
  nkeys := 0;

  LIB$FREE_VM(number_of_bytes := xab.xab$w_lrl,
              base_address := rab.rab$l_ubf);
*)	      
  $CLOSE(fab);
  
END;


FUNCTION SPLIT(s:STRING; at:CHAR; VAR left:STRING; VAR right:STRING):BOOLEAN;
VAR i:INTEGER;
BEGIN
  i:=INDEX(s,at);
  IF i>0 THEN BEGIN
    left:=SUBSTR(s,1,i-1);
    right:=SUBSTR(s,i+1,LENGTH(s)-i);
    SPLIT:=TRUE;
  END ELSE BEGIN
    left:=s;
    right:='';
    SPLIT:=FALSE;
  END;
END;


PROCEDURE TRIM(VAR s:STRING);
CONST
  blanks 	= [CHR(32),CHR(9),CHR(160)]; (* space, tab, nbsp *)
BEGIN
  WHILE s[s.length] IN blanks DO s.length:=s.length-1;
  WHILE s[1] IN blanks DO s:=SUBSTR(s,2,s.length-1);
END;


PROCEDURE DO_SECTION(i: INTEGER);
BEGIN
  PROCESS_PAGE(sects^[i].partno, rab);
END;


PROCEDURE DO_TITLE_FILE_SHELF_SECTION;
VAR
  param:	STRING(255);
  booksection:	STRING(255);
  bookname:	STRING(255);
  bookshelf:	STRING(255);
  booksec,i:	INTEGER;
  debugstr:	STRING(80);
BEGIN
  LIB$GET_FOREIGN(param.body,'File Section URL Shelf:',param.length);
  TRIM(param);
  SPLIT(param,' ',bookname,param);
  book_id:=bookname;
  i:=INDEX(book_id,'.');
  IF i>0 THEN book_id.length:=i-1;
  i:=INDEX(book_id,':');
  IF i>0 THEN book_id:=SUBSTR(book_id,i+1,book_id.length-i);
  i:=INDEX(book_id,']');
  IF i>0 THEN book_id:=SUBSTR(book_id,i+1,book_id.length-i);
  i:=INDEX(book_id,'>');
  IF i>0 THEN book_id:=SUBSTR(book_id,i+1,book_id.length-i);
  SPLIT(param,' ',booksection,param);
  SPLIT(param,' ',book_url,param);
  bookshelf:=param;
  IF bookshelf='' THEN bookshelf:='decw$book:*.decw$bookshelf';
  READV(booksection,booksec);

  IF ODD(LIB$SYS_TRNLOG('HTBOOK_DEBUG',debugstr.length,debugstr.body))
    THEN READV(debugstr,debug,error:=CONTINUE);
  IF INDEX(debugstr,'FONT')>0 THEN debug_font:=TRUE;
  IF INDEX(debugstr,'SPOT')>0 THEN debug_spot:=TRUE;
  IF INDEX(debugstr,'SECMAP')>0 THEN debug_secmap:=TRUE;

  IF ODD(OPEN_BOOK(bookname,'decw$book:.decw$bookshelf')) THEN BEGIN
    IF booksec<0
      THEN DO_SECTION(-booksec)
      ELSE
    IF booksec>0 
      THEN PROCESS_PAGE(booksec, rab)
      ELSE
    PROCESS_PAGE(keypart[1], rab);	      (* show contents table *)	      
    CLOSE_BOOK;
  END;

END;

BEGIN	(********************* PROGRAM ****************************)
  OPEN(output, 'PAS$OUTPUT:', record_length:=2048);	(* longer lines *)
  REWRITE(output);
  INIT_RMS_BLK(fab, rab);
  DO_TITLE_FILE_SHELF_SECTION;
END.

